"""TacoBI is a framework for building and running BI applications."""

from tacobi.bi_app import TacoBIApp
from tacobi.view.view_manager import ViewManager

__all__ = ["TacoBIApp", "ViewManager", "DataSource"]
